/*
 * Copyright by ams AG
 * All rights are reserved.
 *
 * IMPORTANT - PLEASE READ CAREFULLY BEFORE COPYING, INSTALLING OR USING
 * THE SOFTWARE.
 *
 * THIS SOFTWARE IS PROVIDED FOR USE ONLY IN CONJUNCTION WITH AMS PRODUCTS.
 * USE OF THE SOFTWARE IN CONJUNCTION WITH NON-AMS-PRODUCTS IS EXPLICITLY
 * EXCLUDED.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <linux/kernel.h>
#include <linux/i2c.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/string.h>
#include <linux/irq.h>
#include <linux/mutex.h>
#include <linux/unistd.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <linux/input.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/pm.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/string.h>
#include <linux/uaccess.h>
#include <linux/kthread.h>
#include <linux/freezer.h>
#include <linux/gpio/consumer.h>
#include <linux/time.h>

#include <linux/i2c/ams/ams_tsl2585_reg.h>
#include <linux/i2c/ams/ams_tsl2585.h>
#include "ams_i2c.h"
#include "ams_tsl2585_als.h"
#include "ams_tsl2585_fifo.h"
#include "ams_tsl2585_flicker.h"

#ifdef CONFIG_OF
#include <linux/of_device.h>
#endif

#undef TEST
#undef DEBUG

#ifdef TEST
static u8 refcnt;
#endif
/*
 * ID, REVID, AUXID
 */
struct device_ids {
    uint8_t device;
    uint8_t rev;
    uint8_t aux;
};

static struct device_ids const dev_ids[] = {
    { .device = 0x5c,.rev = 0x11,.aux = 0x04 },
};

static char const *dev_names[] = {
    "tsl2585",
};

static struct i2c_device_id tsl2585_idtable[] = {
    { "tsl2585", 0 },
    {}
};
MODULE_DEVICE_TABLE(i2c, tsl2585_idtable);

#define STR1_MAX (16)
#define STR2_MAX (256)


static void tsl2585_smux_init(struct ams_chip *chip)
{

    /*
     *  The sequencer has 4 steps. Sequence steps are performed
     *  in order beginning with step 0. A sequence round is
     *  step0, step1, step2, step3.
     *
     *  Only steps 0 and 1 are used in this driver.
     *  ALS data is collected in step 0.
     *  Flicker data is collected in step 1.
     *  
     *        /-----------------Sequence Round-----------------\
     *            step0       step1       step2       step3   
     *             ALS       FLICKER      UNUSED      UNUSED
     *        -------------------------------------------------
     *        |           |           |           |           |
     *  mod0  | photopic  |  photopic |    n/a    |    n/a    |
     *        |           |           |           |           |
     *        -------------------------------------------------
     *        |           |           |           |           |
     *  mod1  |    ir     |    n/a    |    n/a    |    n/a    |
     *        |           |           |           |           |
     *        -------------------------------------------------
     *        |           |           |           |           |
     *  mod2  |    uv     |    n/a    |    n/a    |    n/a    |
     *        |           |           |           |           |
     *        -------------------------------------------------
     *
     * Photodiode Filter Assignment        
     * pd0 = IR
     * pd1 = PHOTOPIC
     * pd2 = IR
     * pd3 = UV
     * pd4 = UV
     * pd5 = PHOTOPIC
     * 
     * Assign the Photopic photodiodes to modulator 0.
     * Assign the IR photodiodes to modulator 1.
     * Assign the UV photodiodes to modulator 2.
     *
     */
    
    /* -------------- */
    /* --- Step 0 --- */
    /* -------------- */
    /* pd0->M1, pd1->M0, pd2->M1, pd3->M2 */
    ams_i2c_write(chip->client, chip->shadow,
                  TSL2585_REG_MEAS_SEQR_STEP0_MOD_PHDX_SMUX_L, 0xE6);

    /* pd4->M2, pd5->M0 */
    ams_i2c_write(chip->client, chip->shadow,
                  TSL2585_REG_MEAS_SEQR_STEP0_MOD_PHDX_SMUX_H, 0x07);

    
    /* -------------- */
    /* --- Step 1 --- */
    /* -------------- */
    /* pd0->M1, pd1->M0, pd2->M1, pd3->M2 */
    ams_i2c_write(chip->client, chip->shadow,
                  TSL2585_REG_MEAS_SEQR_STEP1_MOD_PHDX_SMUX_L, 0xE6);

    /* pd4->M2, pd5->M0 */
    ams_i2c_write(chip->client, chip->shadow,
                  TSL2585_REG_MEAS_SEQR_STEP1_MOD_PHDX_SMUX_H, 0x07);

    /* Do not write ALS data to fifo */
    ams_i2c_write(chip->client, chip->shadow,
                  TSL2585_REG_MOD_FIFO_DATA_CFG0, 0x0f);
    ams_i2c_write(chip->client, chip->shadow,
                  TSL2585_REG_MOD_FIFO_DATA_CFG1, 0x0f);
    ams_i2c_write(chip->client, chip->shadow,
                  TSL2585_REG_MOD_FIFO_DATA_CFG2, 0x0f);
}

static void tsl2585_sequencer_init(struct ams_chip *chip)
{
    /* Assign ALS to sequencer step 0 */
    ams_i2c_modify(chip->client, chip->shadow,
                   TSL2585_REG_MEAS_SEQR_ALS_FD_1,
                   TSL2585_MASK_MEASUREMENT_SEQUENCER_ALS_PATTERN,
                   0x01);

    /* Assign Flicker to sequencer step 1 */
    ams_i2c_modify(chip->client, chip->shadow,
                   TSL2585_REG_MEAS_SEQR_FD_0,
                   TSL2585_MASK_MEASUREMENT_SEQUENCER_MOD0_FD_PATTERN,
                   0x02);
}

static void tsl2585_set_sample_time(struct ams_chip *chip, uint16_t time)
{
    chip->sample_time = time;

    ams_i2c_write(chip->client, chip->shadow,
                  TSL2585_REG_SAMPLE_TIME0, (uint8_t)((uint16_t)time & 0x00FF));
    ams_i2c_write(chip->client, chip->shadow,
                  TSL2585_REG_SAMPLE_TIME1, (uint8_t)(((uint16_t)time & 0x0700) >> 8));
}

static void tsl2585_set_wait_time(struct ams_chip *chip, uint8_t time)
{
    chip->wait_time = time;

    ams_i2c_write(chip->client, chip->shadow, TSL2585_REG_WTIME, time);

}

static void tsl2585_set_agc_enable(struct ams_chip *chip, uint8_t on_off)
{
    /* Enable / Disable AGC */
    uint8_t reg_value;

    if (on_off) {
        reg_value = TSL2585_MASK_MOD_CALIB_NTH_ITERATION_AGC_ENABLE;
    }
    else {
        reg_value = 0;
    }

    ams_i2c_modify(chip->client, chip->shadow,
                   TSL2585_REG_MOD_CALIB_CFG2,
                   TSL2585_MASK_MOD_CALIB_NTH_ITERATION_AGC_ENABLE,
                   reg_value);

}

static void tsl2585_set_agc_mode(struct ams_chip *chip, uint8_t mode)
{
    uint8_t on_off;
    chip->agc_mode = mode;

    switch (mode) {
    case 1:
        /* Turn on Saturation AGC */
        ams_i2c_modify(chip->client, chip->shadow,
                       TSL2585_REG_MEAS_SEQR_STEP1_MOD_PHDX_SMUX_H,
                       TSL2585_MASK_MEASUREMENT_SEQUENCER_AGC_ASAT_PATTERN,
                       TSL2585_MASK_MEASUREMENT_SEQUENCER_AGC_ASAT_PATTERN);

        /* Turn off Predict AGC */
        ams_i2c_modify(chip->client, chip->shadow,
                       TSL2585_REG_MEAS_SEQR_STEP2_MOD_PHDX_SMUX_H,
                       TSL2585_MASK_MEASUREMENT_SEQUENCER_AGC_PREDICT_PATTERN,
                       0x00);

        on_off = 1;
        break;

    case 2:
        /* Turn on Predict AGC */
        ams_i2c_modify(chip->client, chip->shadow,
                       TSL2585_REG_MEAS_SEQR_STEP2_MOD_PHDX_SMUX_H,
                       TSL2585_MASK_MEASUREMENT_SEQUENCER_AGC_PREDICT_PATTERN,
                       TSL2585_MASK_MEASUREMENT_SEQUENCER_AGC_PREDICT_PATTERN);

        /* Turn off Saturation AGC */
        ams_i2c_modify(chip->client, chip->shadow,
                       TSL2585_REG_MEAS_SEQR_STEP1_MOD_PHDX_SMUX_H,
                       TSL2585_MASK_MEASUREMENT_SEQUENCER_AGC_ASAT_PATTERN,
                       0x00);
        on_off = 1;
        break;

    case 3:
        /* Turn on both Saturation AGC and Predict AGC */
        ams_i2c_modify(chip->client, chip->shadow,
                       TSL2585_REG_MEAS_SEQR_STEP1_MOD_PHDX_SMUX_H,
                       TSL2585_MASK_MEASUREMENT_SEQUENCER_AGC_ASAT_PATTERN,
                       TSL2585_MASK_MEASUREMENT_SEQUENCER_AGC_ASAT_PATTERN);

        ams_i2c_modify(chip->client, chip->shadow,
                       TSL2585_REG_MEAS_SEQR_STEP2_MOD_PHDX_SMUX_H,
                       TSL2585_MASK_MEASUREMENT_SEQUENCER_AGC_PREDICT_PATTERN,
                       TSL2585_MASK_MEASUREMENT_SEQUENCER_AGC_PREDICT_PATTERN);
        on_off = 1;
        break;

    case 0:
    default:
        /* Disable all AGC */
        ams_i2c_modify(chip->client, chip->shadow,
                       TSL2585_REG_MEAS_SEQR_STEP1_MOD_PHDX_SMUX_H,
                       TSL2585_MASK_MEASUREMENT_SEQUENCER_AGC_ASAT_PATTERN,
                       0x00);

        ams_i2c_modify(chip->client, chip->shadow,
                       TSL2585_REG_MEAS_SEQR_STEP2_MOD_PHDX_SMUX_H,
                       TSL2585_MASK_MEASUREMENT_SEQUENCER_AGC_PREDICT_PATTERN,
                       0x00);

        on_off = 0;
        break;


    }

    /* Set for every iteration */
    ams_i2c_modify(chip->client, chip->shadow,
                   TSL2585_REG_MOD_CALIB_CFG0,
                   TSL2585_MASK_MOD_CALIB_NTH_ITERATION,
                   1);


    tsl2585_set_agc_enable(chip, on_off);
}

static void tsl2585_set_agc_num_samples(struct ams_chip *chip, uint16_t num_samples)
{
    chip->agc_num_samples = num_samples;

    ams_i2c_write(chip->client, chip->shadow,
                  TSL2585_REG_AGC_NR_SAMPLES_LO, (uint8_t)((uint16_t)num_samples & 0x00FF));
    ams_i2c_write(chip->client, chip->shadow,
                  TSL2585_REG_AGC_NR_SAMPLES_HI, (uint8_t)(((uint16_t)num_samples & 0x0700) >> 8));
}

void tsl2585_clear_sai(struct ams_chip *chip)
{
    ams_i2c_modify(chip->client, chip->shadow,
                   TSL2585_REG_CONTROL,
                   TSL2585_MASK_CLEAR_SAI_ACTIVE,
                   TSL2585_MASK_CLEAR_SAI_ACTIVE);
}

/*  Attributes  */

static ssize_t all_regs_show(struct device *dev, struct device_attribute *attr,
                             char *buf)
{
    struct ams_chip *chip = dev_get_drvdata(dev);
    uint8_t i;
    uint8_t j;
    int len = 0;
    uint8_t byte = 0;

    AMS_MUTEX_LOCK(&chip->lock);
    for (i = 0x80; i >= 0x80; i += 8) {
        len += scnprintf(buf + len, PAGE_SIZE - len, "0x%02X: ", i);
        for (j = 0; j < 8; ++j) {
            byte = i2c_smbus_read_byte_data(chip->client, i + j);
            len += scnprintf(buf + len, PAGE_SIZE - len, "0x%02x\t\t", byte);
        }
        len += scnprintf(buf + len, PAGE_SIZE - len, "\n");
    }
    AMS_MUTEX_UNLOCK(&chip->lock);
    return len;
}


static ssize_t attribute_store(struct device *dev, struct device_attribute *attr,
                               const char *buf, size_t count);
static ssize_t attribute_show(struct device *dev, struct device_attribute *attr,
                              char *buf);
#define AMS_ATTR(name)                                          \
    DEVICE_ATTR(name, 0644, attribute_show, attribute_store)

static DEVICE_ATTR_RO(all_regs);
static AMS_ATTR(photopic_channel);
static AMS_ATTR(ir_channel);
static AMS_ATTR(flicker_channel);

static struct attribute *dbg_attrs[] = {
    &dev_attr_all_regs.attr,
    NULL,
};

static struct attribute *mux_attrs[] = {
    &dev_attr_photopic_channel.attr,
    &dev_attr_ir_channel.attr,
    &dev_attr_flicker_channel.attr,
    NULL,
};

static ssize_t attribute_store(struct device *dev, struct device_attribute *attr,
                               const char *buf, size_t count)
{
    int ret;
    uint32_t i;
    uint8_t channel;
    struct ams_chip *chip = dev_get_drvdata(dev);

    for (i = 0; i < (sizeof(mux_attrs) / sizeof(struct attribute *) - 1); ++i) {
        if (!strncmp(attr->attr.name, mux_attrs[i]->name, strlen(attr->attr.name))) {
            AMS_MUTEX_LOCK(&chip->lock);
            ret = kstrtol(buf, 0, (long *)(&(channel)));
            if (channel >= 0 && channel <= 2) {
                chip->channels[i] = channel;
            }
            AMS_MUTEX_UNLOCK(&chip->lock);
            return count;
        }
    }

    return count;
}

static ssize_t attribute_show(struct device *dev, struct device_attribute *attr,
                              char *buf)
{
    struct ams_chip *chip = dev_get_drvdata(dev);
    int ret = 0;
    uint32_t i;

    for (i = 0; i < (sizeof(mux_attrs) / sizeof(struct attribute *) - 1); ++i) {
        if (!strncmp(attr->attr.name, mux_attrs[i]->name, strlen(attr->attr.name))) {
            AMS_MUTEX_LOCK(&chip->lock);
            ret = snprintf(buf, PAGE_SIZE, "%d\n", chip->channels[i]);
            AMS_MUTEX_UNLOCK(&chip->lock);
            return ret;
        }
    }
    return ret;
}

static const struct attribute_group mux_attrs_group = {
    .name = "mux",
    .attrs = mux_attrs,
};

static const struct attribute_group dbg_attrs_group = {
    .name = "dbg",
    .attrs = dbg_attrs,
};

static const struct attribute_group *groups[] = {
    &mux_attrs_group,
    &dbg_attrs_group,
    NULL,
};

/*  IRQ Handling    */
static uint8_t get_status(struct ams_chip *chip)
{
    ams_i2c_read(chip->client, TSL2585_REG_STATUS,
                 &chip->shadow[TSL2585_REG_STATUS]);
    return chip->shadow[TSL2585_REG_STATUS];
}

static void irq_callback(struct ams_chip *chip, struct feature_callbacks *callbacks, uint8_t status)
{
    if (callbacks) {
        if (callbacks->irq) {
            callbacks->irq(chip, status);
        }
    }
}

static int irq_handler(struct ams_chip *chip)
{
    uint8_t status;

    AMS_MUTEX_LOCK(&chip->lock);
    status = get_status(chip);
    if (!status) {
        dev_info(&chip->client->dev, "No status to handle\n");
        AMS_MUTEX_UNLOCK(&chip->lock);
        return 0;
    }

    irq_callback(chip, chip->als, status);
    irq_callback(chip, chip->fifo, status);
    irq_callback(chip, chip->flicker, status);

    /* clear the interrupts we processed */
    ams_i2c_write_direct(chip->client, TSL2585_REG_STATUS, status);
    AMS_MUTEX_UNLOCK(&chip->lock);
    return 1;
}

static irqreturn_t tsl2585_irq(int irq, void *handle)
{
    struct ams_chip *chip = handle;
    struct device *dev = &chip->client->dev;

    if (chip->in_suspend) {
        dev_info(dev, "%s: in suspend\n", __func__);
        chip->irq_pending = 1;
    }
    else if (irq_handler(chip)) {
        return IRQ_HANDLED;
    }

    return IRQ_NONE;
}

static int poll_irq_thread(void *context)
{
    struct ams_chip *chip = (struct ams_chip *)context;
    int us_sleep = 0;

    AMS_MUTEX_LOCK(&chip->lock);
    // Poll period is interpreted in units of 100 usec
    us_sleep = chip->platform_data->poll_period * 100;
    dev_info(&chip->client->dev, "Starting tsl2585 irq polling thread, period: %u us\n", us_sleep);
    AMS_MUTEX_UNLOCK(&chip->lock);
    while (!kthread_should_stop()) {
        (void)irq_handler(chip);
        usleep_range(us_sleep, us_sleep + us_sleep / 10);
    }
    return 0;
}

static int init_irq(struct ams_chip *chip)
{
    unsigned long default_irq_trigger;
    int ret;

    if (chip->platform_data->poll_period == 0) {
        /*** Use Interrupt I/O instead of polled ***/
        /***** Setup GPIO IRQ handler *****/
        default_irq_trigger =
            irqd_get_trigger_type(irq_get_irq_data(chip->client->irq));
        ret = devm_request_threaded_irq(&chip->client->dev, chip->client->irq,
                                        NULL, &tsl2585_irq,
                                        default_irq_trigger |
                                        IRQF_SHARED |
                                        IRQF_ONESHOT,
                                        dev_name(&chip->client->dev), chip);
        if (ret) {
            dev_err(&chip->client->dev, "Failed to request irq %d\n", chip->client->irq);
            return -1;
        }
        ams_i2c_write_direct(chip->client, TSL2585_REG_STATUS, 0x00);
    }
    else {
        /*** Use Polled I/O instead of interrupt ***/
        chip->poll_irq = kthread_run(poll_irq_thread,
                                     (void *)chip,
                                     "irq_poll");
        if (IS_ERR(chip->poll_irq)) {
            dev_err(&chip->client->dev, "Error starting IRQ polling thread.\n");
            return -1;
        }
    }

    return 0;
}

/*  Device Initialization   */
static int get_id(struct ams_chip *chip, struct device_ids *ids)
{
    ams_i2c_read(chip->client, TSL2585_REG_AUXID, &ids->aux);
    ams_i2c_read(chip->client, TSL2585_REG_REVID, &ids->rev);
    ams_i2c_read(chip->client, TSL2585_REG_ID, &ids->device);

    return 0;
}

#ifdef CONFIG_OF
#define PROPERTY_READ_u32(name, value, pdata)                   \
    if (!of_property_read_u32(pdata->of_node, #name, &value))   \
        pdata->parameters.name = value;

static int init_device_tree(struct tsl2585_i2c_platform_data *pdata)
{   /* Read from device tree overlay */
    uint32_t val;

    if (!pdata->of_node) {
        return 0;
    }

    if (!of_property_read_u32(pdata->of_node, "poll_x100us_period", &val)) {
        pdata->poll_period = val;
    }


    return 0;
}

static int init_device_tree_general(struct ams_chip *chip)
{
    struct device_node *node = of_get_child_by_name(chip->platform_data->of_node, "general");
    uint32_t val;
    if (node) {
        if (!of_property_read_u32(node, "sample_time", &val)) {
            chip->sample_time = (uint16_t)val;
        }

        if (!of_property_read_u32(node, "wait_time", &val)) {
            chip->wait_time = (uint16_t)val;
        }

        if (!of_property_read_u32(node, "agc_mode", &val)) {
            chip->agc_mode = (uint8_t)val;
        }

        if (!of_property_read_u32(node, "agc_num_samples", &val)) {
            chip->agc_num_samples = (uint16_t)val;
        }

        return 1;
    }

    return 0;
}

static const struct of_device_id of_match[] = {
    { .compatible = "ams,tsl2585" },
    {}
};
MODULE_DEVICE_TABLE(of, of_match);
#endif

static int compare_ids(const struct device_ids *expected, struct device_ids *actual)
{
    if ((expected->device == actual->device)
        && (expected->rev == actual->rev)) {
        return 0;
    }

    return 1;
}

static int ams_tsl2585_init_general(struct ams_chip *chip)
{

    if (chip->input) {

        tsl2585_smux_init(chip);
        tsl2585_sequencer_init(chip);

        if (init_device_tree_general(chip)) {
            tsl2585_set_sample_time(chip, chip->sample_time);
            tsl2585_set_wait_time(chip, chip->wait_time);
            tsl2585_set_agc_num_samples(chip, chip->agc_num_samples);
            tsl2585_set_agc_mode(chip, chip->agc_mode);
            return 0;
        }
    }
    return -ENODEV;
}

static int init_features(struct ams_chip *chip)
{

    if (ams_tsl2585_init_general(chip)) {
        goto als_failed;
    }
    if (ams_tsl2585_init_als(chip)) {
        goto als_failed;
    }
    if (ams_tsl2585_init_flicker(chip)) {
        goto flicker_failed;
    }
    if (ams_tsl2585_init_fifo(chip)) {
        goto fifo_failed;
    }
    if (init_irq(chip)) {
        goto irq_failed;
    }
    return 0;

irq_failed:
    ams_tsl2585_remove_fifo(chip);
fifo_failed:
    ams_tsl2585_remove_flicker(chip);
flicker_failed:
    ams_tsl2585_remove_als(chip);
als_failed:
    return -ENODEV;
}

static void remove_features(struct ams_chip *chip)
{
    ams_tsl2585_remove_fifo(chip);
    ams_tsl2585_remove_flicker(chip);
    ams_tsl2585_remove_als(chip);
}

static int init_input_device(struct ams_chip *chip)
{
    chip->input = devm_input_allocate_device(&chip->client->dev);
    chip->input->name = "tsl2585";
    chip->input->id.bustype = BUS_I2C;
    input_set_drvdata(chip->input, chip);
    if (input_register_device(chip->input)) {
        dev_err(&chip->client->dev, "%s: Can't register input device '%s'\n", __func__, chip->platform_data->name);
        return -ENODEV;
    }
    if (sysfs_create_groups(&chip->input->dev.kobj, groups)) {
        dev_err(&chip->client->dev, "Error creating syfs attribute group.\n");
    }
    return 0;
}

static void power_device(struct ams_chip *chip, uint8_t power)
{   /* Set PON bit */
    if (power) {
        ams_i2c_modify(chip->client, chip->shadow,
                       TSL2585_REG_ENABLE, 0x01, 0x01);
    }
    else {
        ams_i2c_modify(chip->client, chip->shadow,
                       TSL2585_REG_ENABLE, 0x01, 0x00);
    }
}

static int validate_device(struct ams_chip *chip)
{   /* Verify chip IDs and determine device type */
    struct device *dev = &chip->client->dev;
    struct device_ids ids;
    int i;
    int j = ARRAY_SIZE(dev_ids);

    if (!i2c_check_functionality(chip->client->adapter, I2C_FUNC_SMBUS_BYTE_DATA)) {
        dev_err(dev, "%s: i2c smbus byte data unsupported\n", __func__);
        return -EOPNOTSUPP;
    }
    get_id(chip, &ids);
    dev_info(dev, "%s: device id:%02x device revid:%02x device aux:%02x\n",
             __func__, ids.device, ids.rev, ids.aux);
    for (i = 0; i < ARRAY_SIZE(dev_ids); i++) {
        if (!compare_ids(&dev_ids[i], &ids)) {
            j = i;
            break;
        }
    }
    if (j < ARRAY_SIZE(dev_names)) {
        dev_info(dev, "%s: '%s revid_2: 0x%x' detected\n", __func__,
                 dev_names[j], ids.aux);
        chip->device_index = j;
        return 0;
    }
    else {
        dev_err(dev, "%s: chip ID not supported\n", __func__);
        return -EOPNOTSUPP;
    }
}

static int get_platform_data(struct ams_chip *chip)
{   /* Get platform data and initialize */
    struct device *dev = &chip->client->dev;
    struct tsl2585_i2c_platform_data *platform_data = dev->platform_data;

#ifdef CONFIG_OF
    if (!platform_data) {
        platform_data = devm_kzalloc(dev, sizeof(struct tsl2585_i2c_platform_data),
                                     GFP_KERNEL);
        if (!platform_data) {
            dev_info(dev, "still no platform data\n");
            return -ENOMEM;
        }
        if (of_match_device(of_match, &chip->client->dev)) {
            platform_data->of_node = chip->client->dev.of_node;
            init_device_tree(platform_data);
        }
    }
#endif
    chip->platform_data = platform_data;

    return 0;
}

/*  Linux Driver API    */
static int device_probe(struct i2c_client *client,
                        const struct i2c_device_id *idp)
{
    static struct ams_chip *chip;
    struct device *dev = &client->dev;

    chip = devm_kzalloc(dev, sizeof(*chip), GFP_KERNEL);
    if (!chip) {
        return -ENOMEM;
    }
    mutex_init(&chip->lock);
    AMS_MUTEX_LOCK(&chip->lock);
    chip->client = client;
    chip->driver_remove = false;
    if (!validate_device(chip)) {
        if (!get_platform_data(chip)) {
            i2c_set_clientdata(client, chip);
            power_device(chip, 1);
            if (!init_input_device(chip)) {
                if (!init_features(chip)) {
                    AMS_MUTEX_UNLOCK(&chip->lock);
                    return 0;
                }
            }
            i2c_set_clientdata(client, NULL);
        }
        AMS_MUTEX_UNLOCK(&chip->lock);
        return 0;
    }
    dev_err(dev, "Probe failed\n");
    AMS_MUTEX_UNLOCK(&chip->lock);
    return 1;
}

static int device_suspend(struct device *dev)
{
    struct ams_chip *chip = dev_get_drvdata(dev);

    dev_info(dev, "%s\n", __func__);
    AMS_MUTEX_LOCK(&chip->lock);
    chip->in_suspend = 1;
    /* Blocks if work is active */
    ams_tsl2585_remove_fifo(chip);
    if (chip->wake_irq)
        irq_set_irq_wake(chip->client->irq, 1);
    else if (chip->powered)
        dev_info(dev, "powering off\n");
    AMS_MUTEX_UNLOCK(&chip->lock);

    return 0;
}

static int device_resume(struct device *dev)
{
    return 0;
}

static int device_remove(struct i2c_client *client)
{
    struct ams_chip *chip = i2c_get_clientdata(client);

    chip->driver_remove = true;
    wake_up_all(&chip->fifo_wait);
    if (chip->gpiod_interrupt)
        devm_free_irq(&client->dev, client->irq, chip);
    if (chip->platform_data->poll_period != 0)
        (void)kthread_stop(chip->poll_irq);
    if (chip->platform_data->platform_teardown)
        chip->platform_data->platform_teardown(&client->dev);
    i2c_set_clientdata(client, NULL);
    sysfs_remove_groups(&chip->input->dev.kobj, groups);
    remove_features(chip);

    return 0;
}

static const struct dev_pm_ops pm_ops = {
    .suspend = device_suspend,
    .resume = device_resume,
};

static struct i2c_driver ams_driver = {
    .driver = {
        .name = "tsl2585",
        .pm = &pm_ops,
#ifdef CONFIG_OF
        .of_match_table = of_match_ptr(of_match),
#endif
    },
    .id_table = tsl2585_idtable,
    .probe = device_probe,
    .remove = device_remove,
};

module_i2c_driver(ams_driver);

MODULE_AUTHOR("AMS AOS Software<cs.americas@ams.com>");
MODULE_DESCRIPTION("AMS TSL2585 sensor driver");
MODULE_LICENSE("GPL");
MODULE_VERSION("1.7");
