/*
*****************************************************************************
* Copyright by ams AG                                                       *
* All rights are reserved.                                                  *
*                                                                           *
* IMPORTANT - PLEASE READ CAREFULLY BEFORE COPYING, INSTALLING OR USING     *
* THE SOFTWARE.                                                             *
*                                                                           *
* THIS SOFTWARE IS PROVIDED FOR USE ONLY IN CONJUNCTION WITH AMS PRODUCTS.  *
* USE OF THE SOFTWARE IN CONJUNCTION WITH NON-AMS-PRODUCTS IS EXPLICITLY    *
* EXCLUDED.                                                                 *
*                                                                           *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       *
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT         *
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS         *
* FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,     *
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT          *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,     *
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY     *
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT       *
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE     *
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.      *
*****************************************************************************
*/
#include <linux/input.h>
#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/i2c/ams/ams_tsl2585.h>
#include <linux/i2c/ams/ams_tsl2585_reg.h>
#include "ams_i2c.h"
#include <linux/kfifo.h>
#include "ams_tsl2585_als.h"
#include "ams_tsl2585_fifo.h"

void tsl2585_clear_sai(struct ams_chip *chip);

#define STR1_MAX (16)
#define STR2_MAX (256)
#define MAX_FIFO_LEN (512)
#define MAX_NUM_FLICKER_SAMPLES 2048
#define MAX_NUM_FLICKER_SAMPLE_BYTES (MAX_NUM_FLICKER_SAMPLES * 2)

struct fifo_chip {
    struct feature_callbacks callbacks;
    struct work_struct work;
    uint8_t channels;
    uint8_t map;
    int16_t level;
    uint32_t k_len;
    uint32_t overflow;
    uint32_t koverflow;
    uint8_t max;
    uint16_t threshold;
    enum fifo_mode mode;
    struct ams_chip *chip;
};

static uint8_t data[MAX_FIFO_LEN];
static uint8_t fifo_out_data[MAX_NUM_FLICKER_SAMPLE_BYTES];
static DECLARE_KFIFO(ams_fifo, u8, 2 * PAGE_SIZE);

static struct fifo_chip *get_fifo_chip(struct ams_chip *chip)
{
    return (struct fifo_chip *)chip->fifo;
}

void set_fifo_map(struct ams_chip *chip, uint8_t map)
{
    struct fifo_chip *fifo = get_fifo_chip(chip);

    fifo->map = map;

    ams_i2c_modify(chip->client, chip->shadow,
                   TSL2585_REG_MOD_FIFO_DATA_CFG0,
                   TSL2585_MASK_MOD_ALS_FIFO_DATAX_WRITE_ENABLE,
                   0x00);

    ams_i2c_modify(chip->client, chip->shadow,
                   TSL2585_REG_MOD_FIFO_DATA_CFG1,
                   TSL2585_MASK_MOD_ALS_FIFO_DATAX_WRITE_ENABLE,
                   0x00);

    ams_i2c_modify(chip->client, chip->shadow,
                   TSL2585_REG_MOD_FIFO_DATA_CFG2,
                   TSL2585_MASK_MOD_ALS_FIFO_DATAX_WRITE_ENABLE,
                   0x00);
}

static void set_fifo_threshold(struct ams_chip *chip, uint16_t threshold)
{
    struct fifo_chip *fifo = get_fifo_chip(chip);

    fifo->threshold = threshold;

    ams_i2c_write(chip->client, chip->shadow,
                  TSL2585_REG_FIFO_THR, (uint8_t)((threshold & 0x01FF) >> 1));

    ams_i2c_modify(chip->client, chip->shadow,
                   TSL2585_REG_CFG2,
                   TSL2585_MASK_FIFO_THR_LSB,
                   (threshold & 0x0001));
}

static void get_fifo_threshold(struct ams_chip *chip)
{
    struct fifo_chip *fifo = get_fifo_chip(chip);

    ams_i2c_read(chip->client, TSL2585_REG_FIFO_THR, &chip->shadow[TSL2585_REG_FIFO_THR]);
    ams_i2c_read(chip->client, TSL2585_REG_CFG2, &chip->shadow[TSL2585_REG_CFG2]);
    fifo->threshold = (uint16_t)((chip->shadow[TSL2585_REG_FIFO_THR] << 1) | (chip->shadow[TSL2585_REG_CFG2] & 0x01));
}

static void fifo_reset(struct ams_chip *chip)
{
    struct fifo_chip *fifo = get_fifo_chip(chip);
    int16_t level;

    kfifo_reset(&ams_fifo);
    fifo->koverflow = 0;
    fifo->overflow = 0;
    ams_i2c_modify(chip->client, chip->shadow,
                   TSL2585_REG_CONTROL, TSL2585_MASK_FIFO_CLR, TSL2585_MASK_FIFO_CLR);

    ams_i2c_blk_read(chip->client, TSL2585_REG_FIFO_LEVEL,
                     &chip->shadow[TSL2585_REG_FIFO_LEVEL], 2);
    level = (uint16_t)((chip->shadow[TSL2585_REG_FIFO_LEVEL] << 2) |
                       (chip->shadow[TSL2585_REG_FIFO_STATUS0] & 0x03));
    if (level != 0) {
        dev_err(&chip->client->dev, "fifo level not zero: [%d]\n", level);
    }

}

static void configure_fifo(struct ams_chip *chip)
{
    struct fifo_chip *fifo = get_fifo_chip(chip);

    set_fifo_threshold(chip, fifo->threshold);
}

static void enable_fifo(struct ams_chip *chip, uint8_t channels)
{
    struct fifo_chip *fifo = get_fifo_chip(chip);

    if (channels >= 4) {
        set_fifo_map(chip, 0);
        fifo->channels = 1;
    } else {
        set_fifo_map(chip, 0);
        fifo->channels = 0;
        ams_i2c_modify(chip->client, chip->shadow,
                       TSL2585_REG_INTENAB, 0x05, 0x00);

        ams_i2c_modify(chip->client, chip->shadow,
                       TSL2585_REG_SIEN,
                       TSL2585_MASK_SIEN_MEASUREMENT_SEQUENCER,
                       0x00);
        return;
    }
    ams_i2c_modify(chip->client, chip->shadow,
                   TSL2585_REG_SIEN,
                   TSL2585_MASK_SIEN_MEASUREMENT_SEQUENCER,
                   TSL2585_MASK_SIEN_MEASUREMENT_SEQUENCER);

    ams_i2c_modify(chip->client, chip->shadow,
                   TSL2585_REG_INTENAB, 0x05, 0x05); //fifo interrupt

    fifo_reset(chip);
}

static int get_kfifo_len(void)
{
    return kfifo_len(&ams_fifo);
}

ssize_t read_fifo(struct ams_chip *chip, uint16_t *buf, int size)
{
    int len;
    uint16_t tmp;
    int i;
    int j;

    memset(fifo_out_data, 0x0, sizeof(fifo_out_data));
    if (kfifo_len(&ams_fifo) >= size) {
        len = kfifo_out(&ams_fifo, fifo_out_data, size);
        /* byte swap */
        for (i = 0, j = 0; i < size; i+=2, j++) {
            tmp = ((fifo_out_data[i + 1] << 8) | fifo_out_data[i]);
            buf[j] = tmp;
        }
    } else {
        dev_err(&chip->client->dev, "%s: size: %d, fifo_len: %d < %d\n", __func__, size, kfifo_len(&ams_fifo), MAX_NUM_FLICKER_SAMPLE_BYTES);
        len = 0;
    }
    return len;
}

static void fifo_work_cb(struct work_struct *work)
{
    struct fifo_chip *fifo = NULL;

    if (work) {
        fifo = container_of(work, struct fifo_chip, work);
    }
}

static ssize_t fifo_read(struct file *fp, struct kobject *kobj,
                         struct bin_attribute *attr, char *buf, loff_t off, size_t size)
{
    struct device *dev = kobj_to_dev(kobj);
    struct ams_chip *chip = dev_get_drvdata(dev);
    int read;
    int res;
    uint32_t _size = size;

    while ( kfifo_len(&ams_fifo) < _size ) {
        AMS_MUTEX_UNLOCK(&chip->lock);
        res = wait_event_interruptible(chip->fifo_wait,
                                       ((kfifo_len(&ams_fifo) >= _size) ||
                                        chip->driver_remove));
        if (res) return res;
        else if (chip->driver_remove) return 0;
        AMS_MUTEX_LOCK(&chip->lock);
    }

    read = read_fifo(chip, (uint16_t *)buf, _size);
    AMS_MUTEX_UNLOCK(&chip->lock);
    return read;
}

static ssize_t reset_store(struct device *dev,
                           struct device_attribute *attr, const char *buf, size_t count)
{
    struct ams_chip *chip = dev_get_drvdata(dev);

    AMS_MUTEX_LOCK(&chip->lock);
    fifo_reset(chip);
    AMS_MUTEX_UNLOCK(&chip->lock);

    return count;
}

static ssize_t level_show(struct device *dev, struct device_attribute *attr,
                          char *buf)
{
    struct ams_chip *chip = dev_get_drvdata(dev);
    struct fifo_chip *fifo = get_fifo_chip(chip);
    ssize_t ret;

    AMS_MUTEX_LOCK(&chip->lock);
    ret = snprintf(buf, PAGE_SIZE, "%d\n", fifo->level);
    AMS_MUTEX_UNLOCK(&chip->lock);

    return ret;
}

static ssize_t overflowx_show(struct device *dev, struct device_attribute *attr,
                              char *buf)
{
    struct ams_chip *chip = dev_get_drvdata(dev);
    struct fifo_chip *fifo = get_fifo_chip(chip);
    ssize_t ret;

    AMS_MUTEX_LOCK(&chip->lock);
    ret = snprintf(buf, PAGE_SIZE, "%d\n", fifo->overflow);
    AMS_MUTEX_UNLOCK(&chip->lock);

    return ret;
}

static ssize_t k_overflowx_show(struct device *dev, struct device_attribute *attr,
                                char *buf)
{
    struct ams_chip *chip = dev_get_drvdata(dev);
    struct fifo_chip *fifo = get_fifo_chip(chip);
    ssize_t ret;

    AMS_MUTEX_LOCK(&chip->lock);
    ret = snprintf(buf, PAGE_SIZE, "%d\n", fifo->koverflow);
    AMS_MUTEX_UNLOCK(&chip->lock);

    return ret;
}

static ssize_t k_isfull_show(struct device *dev, struct device_attribute *attr,
                             char *buf)
{
    struct ams_chip *chip = dev_get_drvdata(dev);
    struct fifo_chip *fifo = get_fifo_chip(chip);
    ssize_t ret;

    AMS_MUTEX_LOCK(&chip->lock);
    ret = snprintf(buf, PAGE_SIZE, "%d\n", get_kfifo_len() == fifo->max ? 1 : 0);
    AMS_MUTEX_UNLOCK(&chip->lock);

    return ret;
}

static ssize_t k_len_show(struct device *dev, struct device_attribute *attr,
                          char *buf)
{
    struct ams_chip *chip = dev_get_drvdata(dev);
    struct fifo_chip *fifo = get_fifo_chip(chip);
    ssize_t ret;
    int len;

    AMS_MUTEX_LOCK(&chip->lock);
    len = get_kfifo_len();
    ret = snprintf(buf, PAGE_SIZE, "%d\n", len);
    if (len == fifo->max)
        dev_err(&chip->client->dev,
                "kfifo_len == KFIFO_MAX [%d > %d]\n", len, fifo->max);
    AMS_MUTEX_UNLOCK(&chip->lock);

    return ret;
}

static ssize_t threshold_show(struct device *dev, struct device_attribute *attr,
                              char *buf)
{
    struct ams_chip *chip = dev_get_drvdata(dev);
    struct fifo_chip *fifo = get_fifo_chip(chip);
    ssize_t ret;

    AMS_MUTEX_LOCK(&chip->lock);
    get_fifo_threshold(chip);
    ret = snprintf(buf, PAGE_SIZE, "%d\n", fifo->threshold);
    AMS_MUTEX_UNLOCK(&chip->lock);

    return ret;
}

static ssize_t threshold_store(struct device *dev, struct device_attribute *attr,
                               const char *buf, size_t count)
{
    struct ams_chip *chip = dev_get_drvdata(dev);
    u16 threshold = 0;

    (void)sscanf(buf, "%hd", &threshold);

    AMS_MUTEX_LOCK(&chip->lock);
    set_fifo_threshold(chip, threshold);
    AMS_MUTEX_UNLOCK(&chip->lock);
    return count;
}

static ssize_t mode_show(struct device *dev, struct device_attribute *attr,
                         char *buf)
{
    struct ams_chip *chip = dev_get_drvdata(dev);
    struct fifo_chip *fifo = get_fifo_chip(chip);
    ssize_t ret;

    AMS_MUTEX_LOCK(&chip->lock);
//    get_fifo_threshold(chip);
    ret = snprintf(buf, PAGE_SIZE, "%d\n", fifo->mode);
    AMS_MUTEX_UNLOCK(&chip->lock);

    return ret;
}

static ssize_t mode_store(struct device *dev, struct device_attribute *attr,
                          const char *buf, size_t count)
{
    struct ams_chip *chip = dev_get_drvdata(dev);
    struct fifo_chip *fifo = get_fifo_chip(chip);
    u16 mode = 0;

    (void)sscanf(buf, "%hd", &mode);

    AMS_MUTEX_LOCK(&chip->lock);
//    set_fifo_threshold(chip, threshold);
    switch (mode)
    {
    case COMPRESSION_NONE:
    case COMPRESSION_ON:
    case DIFFERENCE_ON:
        fifo->mode = (enum fifo_mode)mode;
        break;
    default:
        break;
    }
    AMS_MUTEX_UNLOCK(&chip->lock);
    return count;
}

static DEVICE_ATTR_WO(reset);
static DEVICE_ATTR_RW(threshold);
static DEVICE_ATTR_RW(mode);
static BIN_ATTR_RO(fifo, PAGE_SIZE);
static DEVICE_ATTR_RO(level);
static DEVICE_ATTR_RO(overflowx);
static DEVICE_ATTR_RO(k_isfull);
static DEVICE_ATTR_RO(k_overflowx);
static DEVICE_ATTR_RO(k_len);

static struct attribute *attrs[] = {
    &dev_attr_reset.attr,
    &dev_attr_level.attr,
    &dev_attr_overflowx.attr,
    &dev_attr_k_isfull.attr,
    &dev_attr_k_overflowx.attr,
    &dev_attr_k_len.attr,
    &dev_attr_threshold.attr,
    &dev_attr_mode.attr,
    NULL,
};

static struct bin_attribute *bin_attrs[] = {
    &bin_attr_fifo,
    NULL,
};

static const struct attribute_group attrs_group = {
    .name = "fifo",
    .attrs = attrs,
    .bin_attrs = bin_attrs,
};

static const struct attribute_group *fifo_groups[] = {
    &attrs_group,
    NULL,
};

static void fifo_callback(struct ams_chip *chip, struct feature_callbacks *callback, uint8_t *data, int size)
{
    if (callback) {
        if (callback->fifo) {
            callback->fifo(chip, data, size);
        }
    }
}

void handle_fifo_data(struct ams_chip *chip, uint8_t *data, int size)
{
    /* Only flicker data in the this version of the driver */
    /* In the future this is where we will
     * include parsing of fifo data.
     * - Multi channel parsing
     * - Decompression (if enabled)
     * - Difference mode (if enabled)
     */
    fifo_callback(chip, chip->flicker, data, size);
}

static void fifo_irq(struct ams_chip *chip, uint8_t status)
{
    int size;
    uint8_t of_val;
    uint8_t uf_val;
    int32_t bytes_read;
    struct fifo_chip *fifo = get_fifo_chip(chip);
    int num_samples = chip->flicker_num_samples + 1;
    int num_sample_bytes = num_samples * 2;
    bool sequence_round_done = false;
    bool sai_active = false;

    /* Check to see if the sequence round is complete, need STATUS4 and STATUS5 */
    if (status & TSL2585_MASK_SINT) {
        ams_i2c_blk_read(chip->client, TSL2585_REG_STATUS4,
                         &chip->shadow[TSL2585_REG_STATUS4], 2);

        if (chip->shadow[TSL2585_REG_STATUS4] &
            TSL2585_MASK_SAI_ACTIVE) {
            sai_active = true;
            /* dev_info(&chip->client->dev, "sai_active\n");             */
        }

        if (chip->shadow[TSL2585_REG_STATUS5] &
            TSL2585_MASK_SINT_MEASUREMENT_SEQUENCER) {
            sequence_round_done = true;
            /* dev_info(&chip->client->dev, "sequence round done\n");             */
        }

    }

    if ((status & TSL2585_MASK_FINT) ||
        /* (status & TSL2585_MASK_SINT) || */
        (true == sai_active) ||
        (true == sequence_round_done)) {
        ams_i2c_blk_read(chip->client, TSL2585_REG_FIFO_LEVEL,
                         &chip->shadow[TSL2585_REG_FIFO_LEVEL], 2);
        fifo->level = (uint16_t)((chip->shadow[TSL2585_REG_FIFO_LEVEL] << 2) |
                                 (chip->shadow[TSL2585_REG_FIFO_STATUS0] & 0x03));

        /* dev_err(&chip->client->dev, */
        /*         "fifo level [%d]\n", fifo->level); */

        of_val = uf_val = chip->shadow[TSL2585_REG_FIFO_STATUS0];
        of_val = (of_val & TSL2585_MASK_FIFO_OV) >> TSL2585_FIFO_OV_SHIFT;
        uf_val = (uf_val & TSL2585_MASK_FIFO_UF) >> TSL2585_FIFO_UF_SHIFT;
        fifo->overflow += of_val;
        if (of_val > 0) {
            dev_err(&chip->client->dev,
                    "fifo overflow [0x%x]\n", of_val);
        }
        if (uf_val > 0) {
            dev_err(&chip->client->dev,
                    "fifo underflow [0x%x]\n", uf_val);
        }
        while (fifo->level > 0) {
            if (fifo->level >= I2C_SMBUS_BLOCK_MAX) {
                size = I2C_SMBUS_BLOCK_MAX;
            } else {
                size = fifo->level;
            }
            memset(&data, 0x0, sizeof(data));
            bytes_read = ams_i2c_blk_read(chip->client, TSL2585_REG_FIFO_DATA, (uint8_t *)&data, size);

            fifo->level -= bytes_read;
            kfifo_in(&ams_fifo, data, bytes_read);
            if (kfifo_is_full(&ams_fifo)) {
                kfifo_reset(&ams_fifo);
                dev_err(&chip->client->dev,
                        "ams_fifo is full, resetting\n");
                break;
            }
        }
        if ((kfifo_len(&ams_fifo) >= num_sample_bytes ) ||
            (true == sai_active) ||
            (true == sequence_round_done)) {
            /* We have enough data for the fft or the sequence round is done. Stop flicker */
            /* dev_info(&chip->client->dev, "fifo samples collected:%d of %d\n", kfifo_len(&ams_fifo), num_sample_bytes); */
            ams_i2c_modify(chip->client, chip->shadow,
                           TSL2585_REG_ENABLE, TSL2585_MASK_FDEN, 0);

            /* Clear the SAI condition */
            if (true == sai_active) {
                tsl2585_clear_sai(chip);
            }

            /* Clear the SINT sequence round done condition */
            if (true == sequence_round_done) {
                ams_i2c_modify(chip->client, chip->shadow,
                               TSL2585_REG_STATUS5,
                               TSL2585_MASK_SINT_MEASUREMENT_SEQUENCER,
                               TSL2585_MASK_SINT_MEASUREMENT_SEQUENCER);
            }

            handle_fifo_data(chip, data, kfifo_len(&ams_fifo));
        }
    }
}

void ams_tsl2585_enable_fifo(struct ams_chip *chip, enum fifo_map map)
{
    configure_fifo(chip);
    enable_fifo(chip, map);
}

ssize_t ams_tsl2585_read_fifo(struct ams_chip *chip, uint16_t *buf, int size)
{
    return read_fifo(chip, buf, size);
}

void ams_tsl2585_fifo_reset(struct ams_chip *chip)
{
    fifo_reset(chip);
}

#define READ_FIFO_PROPERTY(name)                        \
    if (!of_property_read_u32(node, #name, &value)) {   \
        fifo->name = value;                             \
    } else {                                            \
        fifo->name = 0;                                 \
    }

static int init_fifo_device_tree(struct ams_chip *chip)
{
    struct device_node *node = of_get_child_by_name(chip->platform_data->of_node, "fifo");
    struct fifo_chip *fifo = get_fifo_chip(chip);
    int value = 0;

    if (node) {
        READ_FIFO_PROPERTY(threshold)
            READ_FIFO_PROPERTY(mode)
            return 1;
    }
    return 0;
}

void ams_tsl2585_remove_fifo(struct ams_chip *chip)
{
    struct fifo_chip *fifo = get_fifo_chip(chip);

    cancel_work_sync(&fifo->work);
    if (!chip->input) {
        return;
    }
    sysfs_remove_groups(&chip->input->dev.kobj, fifo_groups);
    fifo_reset(chip);
}

int ams_tsl2585_init_fifo(struct ams_chip *chip)
{
    static struct fifo_chip fifo = { {0} };

    fifo.callbacks.irq = fifo_irq;
    chip->fifo = &fifo.callbacks;
    fifo.chip = chip;
    init_waitqueue_head(&chip->fifo_wait);
    if (init_fifo_device_tree(chip)) {
        fifo_reset(chip);
        INIT_WORK(&fifo.work, fifo_work_cb);
        INIT_KFIFO(ams_fifo);
        enable_fifo(chip, FIFO_OFF);
        if (sysfs_create_groups(&chip->input->dev.kobj, fifo_groups)) {
            dev_err(&chip->input->dev, "Error creating sysfs attribute group.\n");
        }
    }
    return 0;
}
