/*
*****************************************************************************
* Copyright by ams AG                                                       *
* All rights are reserved.                                                  *
*                                                                           *
* IMPORTANT - PLEASE READ CAREFULLY BEFORE COPYING, INSTALLING OR USING     *
* THE SOFTWARE.                                                             *
*                                                                           *
* THIS SOFTWARE IS PROVIDED FOR USE ONLY IN CONJUNCTION WITH AMS PRODUCTS.  *
* USE OF THE SOFTWARE IN CONJUNCTION WITH NON-AMS-PRODUCTS IS EXPLICITLY    *
* EXCLUDED.                                                                 *
*                                                                           *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       *
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT         *
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS         *
* FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,     *
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT          *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,     *
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY     *
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT       *
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE     *
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.      *
*****************************************************************************
*/
#include <linux/input.h>
#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/time.h>
#include <linux/i2c/ams/ams_tsl2585.h>
#include <linux/i2c/ams/ams_tsl2585_reg.h>
#include "ams_i2c.h"
#include "ams_tsl2585_flicker.h"
#include "ams_tsl2585_fifo.h"
#include "ams_fft.h"

void tsl2585_clear_sai(struct ams_chip *chip);

#ifdef CONFIG_QUALCOMM_AP
#include <linux/sensors.h>
static struct sensors_classdev flicker_cdev = {
    .name = "tsl2585-flicker",
    .vendor = "AMS",
    .enabled = 0,
    .sensors_enable = NULL,
};
#endif

#define FLICKER_VALUE ABS_MISC
#define FIFO_DEPTH  ABS_VOLUME
#define MAX_NUM_FLICKER_SAMPLES 2048
#define MAX_NUM_FLICKER_SAMPLE_BYTES (MAX_NUM_FLICKER_SAMPLES * 2)

enum flicker_mode {
    SOFTWARE_RAW=1,
    SOFTWARE_FFT=2,
    SOFTWARE_HZ=3
};

struct flicker_chip {
    struct feature_callbacks callbacks;
    struct work_struct work;
    struct input_dev *input;
    uint8_t gain;
    uint16_t rate;
    uint16_t time;
    uint32_t frequency;
    bool enabled;
    enum flicker_mode mode;
    struct ams_chip *chip;
#ifdef CONFIG_QUALCOMM_AP
    struct sensors_classdev cdev;
#endif
};

static struct flicker_chip *get_flicker_chip(struct ams_chip *chip)
{
    return container_of(chip->flicker, struct flicker_chip, callbacks);
}

static void report_flicker(struct ams_chip *chip, int type, int value)
{
    struct flicker_chip *flicker = get_flicker_chip(chip);

    if (flicker->input) {
        input_report_abs(flicker->input, type, value);
        input_sync(flicker->input);
    }
}

static void set_flicker_num_samples(struct ams_chip *chip, uint16_t num_samples)
{
    chip->flicker_num_samples = num_samples;

    ams_i2c_write(chip->client, chip->shadow,
                  TSL2585_REG_FD_NR_SAMPLES0, (uint8_t)((uint16_t)num_samples & 0x00FF));
    ams_i2c_modify(chip->client, chip->shadow,
                   TSL2585_REG_FD_NR_SAMPLES1,
                   TSL2585_MASK_FD_NR_SAMPLES_HI,
                   (uint8_t)(((uint16_t)num_samples & 0x0700) >> 8));
}

static void get_flicker_num_samples(struct ams_chip *chip)
{
    ams_i2c_blk_read(chip->client, TSL2585_REG_FD_NR_SAMPLES0,
                     &chip->shadow[TSL2585_REG_FD_NR_SAMPLES0], 2);

    chip->flicker_num_samples = 0;
    chip->flicker_num_samples = (uint16_t)(*(uint8_t *)&chip->shadow[TSL2585_REG_FD_NR_SAMPLES1] & 0x07) << 8;
    chip->flicker_num_samples |= (uint16_t)*(uint8_t *)&chip->shadow[TSL2585_REG_FD_NR_SAMPLES0];
}

static void configure_flicker(struct ams_chip *chip)
{
    set_flicker_num_samples(chip, chip->flicker_num_samples);

    /* Turn on SAI */
    /* ams_i2c_modify(chip->client, chip->shadow, */
    /*                TSL2585_REG_CFG0, */
    /*                TSL2585_MASK_SAI, */
    /*                TSL2585_MASK_SAI); */

    /* Turn on single shot mode */
    ams_i2c_modify(chip->client, chip->shadow,
                   TSL2585_REG_MEAS_MODE0,
                   TSL2585_MASK_MEASUREMENT_SEQUENCER_SINGLE_SHOT,
                   TSL2585_MASK_MEASUREMENT_SEQUENCER_SINGLE_SHOT);
}

static void enable_flicker(struct ams_chip *chip, uint8_t enable)
{
    struct flicker_chip *flicker = get_flicker_chip(chip);

    if (enable) {
        /* report_flicker(chip, FIFO_DEPTH, 0); */
        /* report_flicker(chip, FLICKER_VALUE, 0); */

        configure_flicker(chip);
        ams_tsl2585_fifo_reset(chip);
        ams_tsl2585_enable_fifo(chip, FIFO_FLICKER);
        ams_i2c_modify(chip->client, chip->shadow,
                       TSL2585_REG_ENABLE, TSL2585_MASK_FDEN, TSL2585_MASK_FDEN);
        flicker->enabled = 1;
    } else {
        ams_i2c_modify(chip->client, chip->shadow,
                       TSL2585_REG_ENABLE, TSL2585_MASK_FDEN, 0x00);
        ams_tsl2585_enable_fifo(chip, FIFO_OFF);

        /* tsl2585_clear_sai(chip); */

        flicker->enabled = 0;
    }
}

static struct timespec old_time = { 0 };

int get_fft(struct ams_chip *chip, uint16_t *out)
{
    ssize_t size;
    static uint16_t buffer[MAX_NUM_FLICKER_SAMPLES] = { 0 };
    int num_samples = chip->flicker_num_samples + 1;
    int num_sample_bytes = num_samples * 2;
    uint16_t max = 0;
    int i;
    
    memset(buffer, 0, MAX_NUM_FLICKER_SAMPLE_BYTES);
    size = ams_tsl2585_read_fifo(chip, buffer, num_sample_bytes);
    if (size < num_sample_bytes) {
        dev_err(&chip->client->dev, "get_fft: %d < %d\n", size, num_sample_bytes);
        return 0 ;
    }
    ams_tsl2585_fifo_reset(chip);
    report_flicker(chip, FIFO_DEPTH, 0);

    ams_i2c_modify(chip->client, chip->shadow,
                   TSL2585_REG_ENABLE, TSL2585_MASK_FDEN, TSL2585_MASK_FDEN);

   /* Normalize to prevent overflow */
    for (i = 0; i < (size / 2); i++)
    {
        if (buffer[i] > max)
        {
            max = buffer[i];
        }
    }
    for (i = 0; i < (size / 2); i++)
    {
        buffer[i] = buffer[i] - (max / 2);
    }
    
    /* dev_info(&chip->client->dev, "calling ams_rfft with %d samples\n", num_samples); */
    if (ams_rfft((int16_t *)buffer, num_samples)) {
        ams_get_magnitude((int16_t *)buffer, out, (num_samples / 2));
        return (num_samples / 2);
    }
    else {
        dev_err(&chip->client->dev, "get_rfft returned 0\n");
    }

    return 0;
}

static uint32_t get_sqrt(uint32_t x)
{
    uint32_t result;
    uint32_t tmp;

    result = 0;
    tmp = (1 << 30);
    while (tmp > x) {
        tmp >>= 2;
    }
    while (tmp != 0) {
        if (x >= (result + tmp)) {
            x -= result + tmp;
            result += 2 * tmp;
        }
        result >>= 1;
        tmp >>= 2;
    }
    return result;
}

static int get_stdev(uint16_t *buff, int mean, int size)
{
    int i;
    uint32_t sum = 0;

    for (i = 0; i < size; ++i) {
        sum += ((buff[i] - mean) * (buff[i] - mean));
    }
    sum = sum / (size - 1);
    return get_sqrt(sum);
}

static int get_mean(uint16_t *buff, int size)
{
    int i;
    int sum = 0;

    for (i = 0; i < size; ++i) {
        sum += buff[i];
    }
    return sum / size;
}

static int get_frequency(struct ams_chip *chip)
{
    struct flicker_chip *flicker = get_flicker_chip(chip);
    static uint16_t buf[MAX_NUM_FLICKER_SAMPLES] = { 0 };
    int max = 0;
    int mean;
    int stdev;
    int i;
    int num_samples = chip->flicker_num_samples + 1;

    memset(buf, 0, sizeof(buf));
    if (get_fft(chip, buf)) {
        buf[0] = 0;
        for (i = 0; i < (num_samples / 2); ++i) {
            if (buf[i] > buf[max]) {
                max = i;
            }
        }
        /* dev_info(&chip->client->dev, "max %d\n", max); */
        mean = get_mean(buf, (num_samples / 2));
        stdev = get_stdev(buf, mean, (num_samples / 2));
        if (buf[max] > (mean + (stdev * 6))) {
            return ((max * flicker->frequency) / num_samples);
        }
    } else {
        dev_err(&chip->client->dev, "get_fft returned 0\n");
    }

    return 0;
}

static ssize_t enable_store(struct device *dev, struct device_attribute *attr,
                            const char *buf, size_t count)
{
    struct ams_chip *chip = dev_get_drvdata(dev);
    ssize_t ret;
    uint8_t enable;

    ret = kstrtol(buf, 0, (long *)(&(enable)));
    if (ret != 0)
        dev_err(&chip->client->dev, "kstrtol() error.\n");
    enable_flicker(chip, enable);

    return count;
}

static ssize_t enable_show(struct device *dev, struct device_attribute *attr,
                           char *buf)
{
    struct ams_chip *chip = dev_get_drvdata(dev);
    struct flicker_chip *flicker = get_flicker_chip(chip);
    int ret;

    AMS_MUTEX_LOCK(&chip->lock);
    ret = snprintf(buf, PAGE_SIZE, "%d\n", flicker->enabled);
    AMS_MUTEX_UNLOCK(&chip->lock);

    return ret;
}

static ssize_t fft_read(struct file *fp, struct kobject *kobj,
                        struct bin_attribute *attr, char *buf, loff_t off, size_t size)
{
    struct device *dev = kobj_to_dev(kobj);
    struct ams_chip *chip = dev_get_drvdata(dev);
    int read;

    if (size < (sizeof(uint16_t) * 1024)) {
        dev_info(dev, "%s: Size %zd is less than size of fft\n", __func__, size);
        return -EAGAIN;
    }
    AMS_MUTEX_LOCK(&chip->lock);
    read = get_fft(chip, (uint16_t *)buf);
    AMS_MUTEX_UNLOCK(&chip->lock);
    return read;
}

static ssize_t raw_read(struct file *fp, struct kobject *kobj,
                        struct bin_attribute *attr, char *buf, loff_t off, size_t size)
{
    struct device *dev = kobj_to_dev(kobj);
    struct ams_chip *chip = dev_get_drvdata(dev);
    int read;

    AMS_MUTEX_LOCK(&chip->lock);
    read = ams_tsl2585_read_fifo(chip, (uint16_t *)buf, size);
    AMS_MUTEX_UNLOCK(&chip->lock);

    return read;
}

static ssize_t frequency_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    int ret = 0;
    struct ams_chip *chip = dev_get_drvdata(dev);
    int frequency;

    AMS_MUTEX_LOCK(&chip->lock);
    frequency = get_frequency(chip);
    ret = snprintf(buf, PAGE_SIZE, "%d\n", frequency);
    AMS_MUTEX_UNLOCK(&chip->lock);
    return ret;
}
static ssize_t config_show(struct device *dev, struct device_attribute *attr,
                           char *buf);
static ssize_t config_store(struct device *dev, struct device_attribute *attr,
                            const char *buf, size_t count);
#define FLICKER_CFG_ATTR(name)                          \
    DEVICE_ATTR(name, 0644, config_show, config_store)

static FLICKER_CFG_ATTR(num_samples);
static FLICKER_CFG_ATTR(mode);
static DEVICE_ATTR_RW(enable);
static BIN_ATTR_RO(fft, PAGE_SIZE);
static BIN_ATTR_RO(raw, PAGE_SIZE);
static DEVICE_ATTR_RO(frequency);

static struct attribute *config_attrs[] = {
    &dev_attr_num_samples.attr,
    &dev_attr_mode.attr,
    &dev_attr_enable.attr,
    NULL,
};

enum flicker_config {
    NUM_FD_SAMPLES_CFG=0,
    MODE_CFG,
};

static ssize_t config_show(struct device *dev, struct device_attribute *attr,
                           char *buf)
{
    struct ams_chip *chip = dev_get_drvdata(dev);
    struct flicker_chip *flicker = get_flicker_chip(chip);
    int ret = 0;
    uint32_t i;

    for (i = 0; i < (sizeof(config_attrs) / sizeof(struct attribute *) - 1); ++i) {
        if (!strncmp(attr->attr.name, config_attrs[i]->name, strlen(attr->attr.name))) {
            AMS_MUTEX_LOCK(&chip->lock);
            switch (i) {
            case NUM_FD_SAMPLES_CFG:
                get_flicker_num_samples(chip);
                ret = snprintf(buf, PAGE_SIZE, "%d\n", chip->flicker_num_samples);
                break;
            case MODE_CFG:
                ret = snprintf(buf, PAGE_SIZE, "%d\n", flicker->mode);
                break;
            default:
                ret = snprintf(buf, PAGE_SIZE, "not found\n");
                break;
            }
            AMS_MUTEX_UNLOCK(&chip->lock);
            return ret;
        }
    }
    return ret;
}

static ssize_t config_store(struct device *dev, struct device_attribute *attr,
                            const char *buf, size_t count)
{
    struct ams_chip *chip = dev_get_drvdata(dev);
    struct flicker_chip *flicker = get_flicker_chip(chip);
    uint32_t i = 0;
    int data;

    (void)sscanf(buf, "%d", (int*)&data);
    for (i = 0; i < (sizeof(config_attrs) / sizeof(struct attribute *) - 1); ++i) {
        if (!strncmp(attr->attr.name, config_attrs[i]->name, strlen(attr->attr.name))) {
            AMS_MUTEX_LOCK(&chip->lock);
            switch (i) {
            case NUM_FD_SAMPLES_CFG:
                set_flicker_num_samples(chip, (uint16_t)data);
                break;
            case MODE_CFG:
                if (data >= 0 && data <= 3) {
                    flicker->mode = data;
                }
                break;
            default:
                break;
            }
            AMS_MUTEX_UNLOCK(&chip->lock);
            return count;
        }
    }
    return count;
}

static const struct attribute_group config_group = {
    .name = "config",
    .attrs = config_attrs,
};

static struct attribute *data_attrs[] = {
    &dev_attr_frequency.attr,
    NULL,
};

static struct bin_attribute *data_bin_attrs[] = {
    &bin_attr_fft,
    &bin_attr_raw,
    NULL,
};

static struct attribute_group data_group = {
    .name = "data",
    .attrs = data_attrs,
    .bin_attrs = data_bin_attrs,
};

static const struct attribute_group *flicker_groups[] = {
    &config_group,
    &data_group,
    NULL,
};

static void flicker_work_cb(struct work_struct *work)
{
    struct ams_chip *chip = NULL;
    struct flicker_chip *flicker = NULL;
    int frequency;
    int num_samples = 0;

    if (work) {
        flicker = container_of(work, struct flicker_chip, work);
        if (flicker) {
            chip = flicker->chip;
            num_samples = chip->flicker_num_samples + 1;
        }
    }
    if (flicker->input) {
        switch (flicker->mode) {
        case SOFTWARE_RAW:
        case SOFTWARE_FFT:
            /* dev_info(&chip->client->dev, "SOFTWARE_RAW:\n"); */
            report_flicker(chip, FIFO_DEPTH, num_samples);
            break;
        case SOFTWARE_HZ:
            frequency = get_frequency(chip);
            if (frequency) {
                /* dev_info(&chip->client->dev, "SOFTWARE_HZ: %d\n", frequency); */
                report_flicker(chip, FLICKER_VALUE, frequency);
            }
            else
            {
                dev_info(&chip->client->dev, "SOFTWARE_HZ: %d\n", frequency);
            }
            break;
        default:
            dev_err(&chip->client->dev, "invalid flicker mode %d\n", flicker->mode);
            break;
        }
    }
}

static void flicker_fifo(struct ams_chip *chip, uint8_t *data, int size)
{
    struct flicker_chip *flicker = get_flicker_chip(chip);
    int num_samples = chip->flicker_num_samples + 1;

    if ((size >= num_samples) &&
        (flicker->mode == SOFTWARE_HZ)) {

        /* temporary hard code */
        flicker->frequency = 4000;
        schedule_work(&flicker->work);
    } else if ((size >= num_samples) &&
               (flicker->mode == SOFTWARE_RAW)) {

        report_flicker(chip, FIFO_DEPTH, size);
        wake_up_interruptible(&chip->fifo_wait);
    }
}


static int flicker_open(struct input_dev *input)
{
    struct ams_chip *chip = dev_get_drvdata(&input->dev);

    report_flicker(chip, FIFO_DEPTH, 0);
    report_flicker(chip, FLICKER_VALUE, 0);
    dev_info(&chip->client->dev, "opening flicker\n");
    enable_flicker(chip, 1);
    getnstimeofday(&old_time);

    return 0;
}

static void flicker_close(struct input_dev *input)
{
    struct ams_chip *chip = dev_get_drvdata(&input->dev);
    struct flicker_chip *flicker = get_flicker_chip(chip);

    cancel_work_sync(&flicker->work);
    dev_info(&chip->client->dev, "closing flicker\n");
    enable_flicker(chip, 0);
}

#ifdef CONFIG_QUALCOMM_AP
static int flicker_set_enable(struct sensors_classdev *sensors_cdev, unsigned int enable)
{
    struct flicker_chip *flicker = container_of(sensors_cdev, struct flicker_chip, cdev);

    if (enable) {
        flicker->input->open(flicker->input);
    } else {
        flicker->input->close(flicker->input);
    }

    return 0;
}
#endif

static int init_flicker_input(struct ams_chip *chip)
{
    struct flicker_chip *flicker = get_flicker_chip(chip);

    flicker->input = devm_input_allocate_device(&chip->client->dev);
    if (!flicker->input) {
        return -ENODEV;
    }
    flicker->input->name = "flicker";
    flicker->input->id.bustype = BUS_I2C;
    set_bit(EV_ABS, flicker->input->evbit);
    set_bit(FLICKER_VALUE, flicker->input->absbit);
    set_bit(FIFO_DEPTH, flicker->input->absbit);
    input_set_abs_params(flicker->input, FLICKER_VALUE, 0, 65535, 0, 0);
    input_set_abs_params(flicker->input, FIFO_DEPTH, 0, 65535, 0, 0);
    flicker->input->open = flicker_open;
    flicker->input->close = flicker_close;
    input_set_drvdata(flicker->input, chip);
    if (input_register_device(flicker->input)) {
        dev_err(&chip->client->dev, "%s: Can't register input device '%s'\n", __func__, flicker->input->name);
        return -ENODEV;
    }
    if (sysfs_create_groups(&flicker->input->dev.kobj, flicker_groups)) {
        dev_err(&chip->client->dev, "Error creating sysfs attribute group.\n");
    }
#ifdef CONFIG_QUALCOMM_AP
    flicker->cdev = flicker_cdev;
    flicker->cdev.sensors_enable = flicker_set_enable;
    if (sensors_classdev_register(&flicker->input->dev, &flicker->cdev)) {
        dev_err(&chip->client->dev, "sensors class register failed\n");
        return -ENODEV;
    }
#endif
    return 0;
}

#define READ_FLICKER_PROPERTY(name)                     \
    if (!of_property_read_u32(node, #name, &value)) {   \
        flicker->name = value;                          \
    } else {                                            \
        flicker->name = 0;                              \
    }

static int init_flicker_device_tree(struct ams_chip *chip)
{
    struct device_node *node = of_get_child_by_name(chip->platform_data->of_node, "flicker");
    struct flicker_chip *flicker = get_flicker_chip(chip);
    int value = 0;

    if (node) {
        if (!of_property_read_u32(node, "num_samples", &value)) {
            chip->flicker_num_samples = value;
        } else {
            chip->flicker_num_samples = 0;
        }

//        READ_FLICKER_PROPERTY(num_samples)
        READ_FLICKER_PROPERTY(mode)
            return 1;
    }
    return 0;
}

void ams_tsl2585_remove_flicker(struct ams_chip *chip)
{
    struct flicker_chip *flicker = get_flicker_chip(chip);

    cancel_work_sync(&flicker->work);
    if (flicker->input) {
        sysfs_remove_link(&chip->input->dev.kobj, "flicker");
        sysfs_remove_groups(&flicker->input->dev.kobj, flicker_groups);
        input_unregister_device(flicker->input);
    }
}

int ams_tsl2585_init_flicker(struct ams_chip *chip)
{
    static struct flicker_chip flicker = { {0} };

    if (chip->input) {
        flicker.chip = chip;
        chip->flicker = &flicker.callbacks;
        if (init_flicker_device_tree(chip)) {
            flicker.callbacks.irq = NULL;
            if (init_flicker_input(chip)) {
                return -ENODEV;
            }
            if (sysfs_create_link(&chip->input->dev.kobj, &flicker.input->dev.kobj, "flicker")) {
                //do nothing
            }
            flicker.callbacks.fifo = flicker_fifo;
            INIT_WORK(&flicker.work, flicker_work_cb);
        }
        return 0;
    }
    return -ENODEV;
}
