/*
*****************************************************************************
* Copyright by ams AG                                                       *
* All rights are reserved.                                                  *
*                                                                           *
* IMPORTANT - PLEASE READ CAREFULLY BEFORE COPYING, INSTALLING OR USING     *
* THE SOFTWARE.                                                             *
*                                                                           *
* THIS SOFTWARE IS PROVIDED FOR USE ONLY IN CONJUNCTION WITH AMS PRODUCTS.  *
* USE OF THE SOFTWARE IN CONJUNCTION WITH NON-AMS-PRODUCTS IS EXPLICITLY    *
* EXCLUDED.                                                                 *
*                                                                           *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       *
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT         *
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS         *
* FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,     *
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT          *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,     *
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY     *
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT       *
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE     *
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.      *
*****************************************************************************
*/

/*! \file
* \brief Device driver for monitoring ambient light intensity in (lux)
* and flicker frequency in the AMS TSL2585.
*/

#ifndef __TSL2585_H
#define __TSL2585_H

#include <linux/types.h>
#include <linux/mutex.h>
#include <linux/workqueue.h>
#include <linux/wait.h>

#ifdef AMS_MUTEX_DEBUG
#define AMS_MUTEX_LOCK(m) { \
        printk(KERN_INFO "%s: Mutex Lock\n", __func__); \
        mutex_lock(m); \
    }
#define AMS_MUTEX_UNLOCK(m) { \
        printk(KERN_INFO "%s: Mutex Unlock\n", __func__); \
        mutex_unlock(m); \
    }
#else
#define AMS_MUTEX_LOCK(m) { \
        mutex_lock(m); \
    }
#define AMS_MUTEX_UNLOCK(m) { \
        mutex_unlock(m); \
    }
#endif

/* Default Params */

#define COEF_SCALE_DEFAULT 1000
#define DGF_DEFAULT        898


// fraction out of 10
#define TENTH_FRACTION_OF_VAL(v, x) ({ \
  int __frac = v; \
  if (((x) > 0) && ((x) < 10)) __frac = (__frac*(x)) / 10 ; \
  __frac; \
})


#define MAX_REGS 256
struct device;

enum tsl2585_pwr_state {
    POWER_ON,
    POWER_OFF,
    POWER_STANDBY,
};

#define INTEGRATION_CYCLE 1390
#define AW_TIME_MS(p)  ((((p) * 1000) + (INTEGRATION_CYCLE - 1)) / INTEGRATION_CYCLE)
#define ALS_PERSIST(p) (((p) & 0xf) << 0)

// lux
#define INDOOR_LUX_TRIGGER    6000
#define OUTDOOR_LUX_TRIGGER    10000
#define TSL2585_MAX_LUX        0xffff
#define TSL2585_MAX_ALS_VALUE    0xffff
#define TSL2585_MIN_ALS_VALUE    10

enum tsl2585_modulator_id {
    TSL2585_MOD_0=0x1,
    TSL2585_MOD_1=0x2,
    TSL2585_MOD_2=0x4,
    TSL2585_MOD_ALL=0x7,
};

struct tsl2585_parameters {
    /* Common */
    u8 single_input;
};

struct als_channels {
    int32_t photopic;
    uint8_t photopic_gain;
    int32_t ir;
    uint8_t ir_gain;
    int32_t uv;
    uint8_t uv_gain;
    uint8_t als_valid;
};

struct ams_chip {
    struct mutex lock;
    struct i2c_client *client;
    struct gpio_desc *gpiod_interrupt;
    struct tsl2585_parameters params;
    struct tsl2585_i2c_platform_data *platform_data;
    struct input_dev *input;
    struct feature_callbacks *als;
    struct feature_callbacks *flicker;
    struct feature_callbacks *fifo;
    struct als_channels channel_data;
    wait_queue_head_t fifo_wait;
    wait_queue_head_t data_wait;
    uint32_t scaled_modulator_clock;
    uint32_t frequency;
    uint16_t sample_time;
    uint8_t  wait_time;
    uint8_t  agc_mode;
    uint16_t agc_num_samples;
    uint16_t flicker_num_samples;
    uint8_t shadow[MAX_REGS];
    uint8_t channel_config;
    uint8_t channels[7];
    int in_suspend;
    int wake_irq;
    int irq_pending;
    bool powered;
    bool amsIndoorMode;
    bool in_calib;
    bool agc_disable;
    bool use_residuals;
    bool driver_remove;
    struct task_struct *poll_irq;
    int brightness;
    uint8_t device_index;
};

struct feature_callbacks {
    void (*irq)(struct ams_chip *chip, uint8_t status);
    void (*fifo)(struct ams_chip *chip, uint8_t *data, int size);
};

// Must match definition in ../arch file
struct tsl2585_i2c_platform_data {
    /* The following callback for power events received and handled by
       the driver.  Currently only for SUSPEND and RESUME */
    int (*platform_power)(struct device *dev, enum tsl2585_pwr_state state);
    int (*platform_init)(void);
    void (*platform_teardown)(struct device *dev);

    char const *name;
    struct tsl2585_parameters parameters;
    bool als_can_wake;
    int poll_period;
#ifdef CONFIG_OF
      struct device_node  *of_node;
#endif
};

#endif /* __TSL2585_H */
