/*
*****************************************************************************
* Copyright by ams AG                                                       *
* All rights are reserved.                                                  *
*                                                                           *
* IMPORTANT - PLEASE READ CAREFULLY BEFORE COPYING, INSTALLING OR USING     *
* THE SOFTWARE.                                                             *
*                                                                           *
* THIS SOFTWARE IS PROVIDED FOR USE ONLY IN CONJUNCTION WITH AMS PRODUCTS.  *
* USE OF THE SOFTWARE IN CONJUNCTION WITH NON-AMS-PRODUCTS IS EXPLICITLY    *
* EXCLUDED.                                                                 *
*                                                                           *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       *
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT         *
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS         *
* FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,     *
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT          *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,     *
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY     *
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT       *
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE     *
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.      *
*****************************************************************************
*/

/*
 * Userapp for TCS3400 to get lux and CCT
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <stdint.h>
#include <inttypes.h>
#include <errno.h>
#include <sys/stat.h>
#include <dirent.h>
#include <sys/select.h>
#include <poll.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <pthread.h>
#include <sys/time.h>
#include <string.h>
#include <signal.h>
#include <syslog.h>

#define EXP_PARAMS 0
#define ARRAY_SIZE(arr) (sizeof(arr)/sizeof(arr[0]))
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define INTEGRATION_CYCLE 2780
#define AW_TIME_MS(p)  ((((p) * 1000) + (INTEGRATION_CYCLE - 1)) / INTEGRATION_CYCLE)
#define ATIME_TO_MAX_ALS(ms) MIN(((AW_TIME_MS(ms))*1024 + 1023), 0xFFFF)

struct als_channels {
    int32_t photopic;
    uint8_t photopic_gain;
    int32_t ir;
    uint8_t ir_gain;
    int32_t uv;
    uint8_t uv_gain;
    uint8_t als_valid;
};
static struct als_channels channel_data;

static const char * dev_name = "als";
static const char * input_sys_path = "/sys/class/input";
static const char * enable_fname = "config/enable";
static const char * raw_fname    = "data/raw";
//
static char sys_path[PATH_MAX] = {0};
static uint32_t verbose = 0;

int find_dev_sysfs_path(const char * dirpath, const char * dev_name,
			char * sys_path, size_t size);
int sysfs_read(const char * sys_entry, char * buf, size_t size);
int sysfs_write_u32(const char * sys_entry, uint32_t val);
int compute_lux();
int compute_uvindex();

int main(int argc, char * argv[])
{
	int c;
	uint32_t rc = 0;
	uint32_t enabled = 0;
	uint32_t raw_data[6];
	char sys_entry[PATH_MAX] = {0};
	char buf[100] = {0};

	while ((c = getopt (argc, argv, "hvg:t:")) != -1) {
		switch (c) {
			case 'h':
				printf("\n\n");
				printf("Help: ./%s [options]\n", __FILE__);
				printf("\n\n");
				printf("Options:\n");
				printf("\t-h: help screen\n");
				printf("\t-v: verbose printing\n");
				printf("\n\n");
				exit(0);
				break;
			case 'v':
				verbose = 1;
				break;
			default:
				fprintf(stderr, "Error parsing cmdline args.\n");
				exit(1);
		}
	}

	if ((argc - 1) < EXP_PARAMS) {
		fprintf(stderr, "Error, expected \'%d\' parameters "
				"only \'%d\' received\n",
				EXP_PARAMS, (argc - 1));
		exit(1);
	}

	// -- search for input event device
	// -- enable als
	// -- read raw values
	rc = find_dev_sysfs_path(input_sys_path, dev_name,
			sys_path, sizeof(sys_path));
	if (rc) {
		fprintf(stderr, "Error (%d), '%s' not found in device directory '%s'. "
				"Is the driver module probed?\n", rc, dev_name, input_sys_path);
		exit(rc);
	}

	// Read Enable state
	snprintf(sys_entry, sizeof(sys_entry),
		"%s/%s", sys_path, enable_fname);
	rc = sysfs_read(sys_entry, buf, sizeof(buf));
	if (rc) {
		fprintf(stderr, "Error (%d), when accessing sysfs_entry.\n", rc);
		exit(rc);
	}
	enabled = strtoul(buf, NULL, 0);
	if (enabled == 0) {
		// Enable ALS
		enabled = 1;
		snprintf(sys_entry, sizeof(sys_entry),
			"%s/%s", sys_path, enable_fname);
		rc = sysfs_write_u32(sys_entry, enabled);
		if (rc) {
			fprintf(stderr, "Error (%d), when accessing sysfs_entry.\n", rc);
			exit(rc);
		}
	}

	//let a measurement complete
	sleep(1);

	// Get raw data
	snprintf(sys_entry, sizeof(sys_entry),
		"%s/%s", sys_path, raw_fname);
	rc = sysfs_read(sys_entry, buf, sizeof(buf));
	if (rc) {
		fprintf(stderr, "Error (%d), when accessing sysfs_entry.\n", rc);
		exit(rc);
	}

	// Parse the raw data
	int cnt=0;
	char delim[] = " ";
	char *ptr = strtok(buf, delim);
	while(ptr != NULL)
	{
                raw_data[cnt] = strtoul(ptr, NULL, 0);
		ptr = strtok(NULL, delim);
		cnt++;
	}
	// save the raw data
        channel_data.photopic      = raw_data[0];
        channel_data.photopic_gain = raw_data[1];
        channel_data.ir            = raw_data[2];
        channel_data.ir_gain       = raw_data[3];
        channel_data.uv            = raw_data[4];
        channel_data.uv_gain       = raw_data[5];
        //channel_data.als_valid;

	printf( "*****************************************************************************************************\n" );
	printf( "*****************************************************************************************************\n" );
	fprintf(stderr, "Raw data photopic = %d ir_data = %d uv_data = %d \n", channel_data.photopic, channel_data.ir, channel_data.uv);
	fprintf(stderr, "Raw data ph gain = %d ir gain = %d uv gain = %d \n", channel_data.photopic_gain, channel_data.ir_gain, channel_data.uv_gain);

	compute_lux();
	compute_uvindex();

	exit(0);
}

int find_dev_sysfs_path(const char * dirpath, const char * dev_name,
			char * sys_path, size_t size)
{
	static const char * dev_name_filename = "name";
	DIR *dir;
	struct dirent *entry;
	char query_name[PATH_MAX] = {0};
	char query_read[100] = {0};
	FILE *fp;

	if (!(dir = opendir(dirpath))) {
		fprintf(stderr, "Error, could not open directory '%s'\n", dirpath);
		return -1;
	}

	while ((entry = readdir(dir)) != NULL) {
		memset(sys_path, 0, size);
		//fprintf(stderr, "Try to open dir '%s'\n", dirpath);
		if (entry->d_type == DT_LNK) {
			if (strcmp(entry->d_name, ".") == 0 ||
			    strcmp(entry->d_name, "..") == 0)
				continue;
			snprintf(sys_path, size, "%s/%s", dirpath, entry->d_name);
			snprintf(query_name, sizeof(query_name),
				"%s/%s", sys_path, dev_name_filename);
			fp = fopen(query_name, "r");
			if (fp == NULL)
				continue;
			if (NULL == fgets(query_read, sizeof(query_read), fp)) {
				fclose(fp);
				continue;
			}
			fclose(fp);
			query_read[strcspn(query_read, "\n")] = 0;
			if (strcmp(query_read, dev_name) == 0) {
				closedir(dir);
				return 0;
			}
		} else {
			// Not a directory
			continue;
		}
	}
	closedir(dir);
	memset(sys_path, 0, size);
	return -1; //not found
}

int sysfs_read(const char * sys_entry, char * buf, size_t size)
{
	FILE *fp;
	memset(buf, 0, size);
	fp = fopen(sys_entry, "r");
	if (ferror(fp) || fp == NULL) {
		fprintf(stderr, "Error opening '%s' file\n", sys_entry);
		return -1;
	} else {
		if (fgets(buf, size, fp) == NULL) {
			fprintf(stderr, "Error reading '%s' file", sys_entry);
			fclose(fp);
			return -1;
		}
		buf[strcspn(buf, "\n")] = 0; //strip newline from output
		if (verbose) {
			printf("read '%s' from '%s'\n", buf, sys_entry);
		}
	}
	fclose(fp);
	return 0;
}

int sysfs_write_u32(const char * sys_entry, uint32_t val)
{
	FILE *fp;
	fp = fopen(sys_entry, "w");
	if (ferror(fp) || fp == NULL) {
		fprintf(stderr, "Error opening '%s' file\n", sys_entry);
		return -1;
	} else {
		if (fprintf(fp, "%u", val) < 0) {
			fprintf(stderr, "Error writing '%s' file\n", sys_entry);
			fclose(fp);
			return -1;
		}
		if (verbose) {
			printf("write '%u' to '%s'\n", val, sys_entry);
		}
	}
	fclose(fp);
	return 0;
}

/**********************************************************************************/
/* Lux = DGFn*((CoefAn*PHO)+(CoefBn*IR))/(ATime*AGain)                            */
/**********************************************************************************/
int compute_lux(void)
{
  float n1_coefA = 0.598;
  float n1_coefB = 0.598;
  float n1_DFG   = 29.915;

  float n2_coefA = 0.546;
  float n2_coefB = -0.028;
  float n2_DFG   = 27.278;
  float again = 2048;
  float atime = 99;

  float coefA;
  float coefB;
  float DGF;

  int ph_gain;
  int ir_gain;

  float photopic_data;
  float ir_data;

  float gain_ratio;
  float ir_pho_ratio;
  double lux = 0;
  float cpl;

  printf( "compute_lux \n" );

  /* raw data for lux*/
  photopic_data = channel_data.photopic;
  ir_data = channel_data.ir;

  ph_gain =  channel_data.photopic_gain;
  ir_gain =  channel_data.ir_gain;

  /* Normalize ir data if gains are different */
  if (ph_gain != ir_gain)
  {
    gain_ratio = ir_gain / ph_gain;
    ir_data = ir_data / gain_ratio;
  }

  if (photopic_data  == 0)
  {
    return 0;
  }

  /* IR/PHO < 1.042 find the ir/pho ratio */
  ir_pho_ratio = ir_data/photopic_data;
  if (ir_pho_ratio  < 1.042)
  {
    coefA = n1_coefA;
    coefB = n1_coefB;
    DGF   = n1_DFG;
    printf("n=1 coefficient -> coefA: %f, coefB: %f, DGF: %f \n", coefA, coefB, DGF );
  }
  else
  {
    coefA = n2_coefA;
    coefB = n2_coefB;
    DGF   = n2_DFG;
    printf("n=2 coefficient -> coefA: %f, coefB: %f, DGF: %f \n", coefA, coefB, DGF );
  }

  /* atime * again */
  cpl = atime * again;
  lux = (DGF * ((photopic_data * coefA) + (ir_data * coefB))) / cpl;

  /* Make sure no negative value */
  if (lux > 0)
  {
    printf("lux with boot,diffuser ->                  lux: %f \n", lux);
    return lux;
  }
  else
  {
    printf("lux with boot,diffuser -> zero or negative lux: %f \n", lux);
    return 0;
  }

  //  printf( "compute_lux\n" );
  printf( "\n" );
  return lux;
}

/**********************************************************************************/
/**********************************************************************************/
int compute_uvindex(void)
{
  float uv_coef1 = 0.000040;
  float uv_coef2 = 0.000021;
  float uv_coef3 = 0.001048;
  float uv_coef4 = 2.561744;

  float uv_dcoef1 = 0.002201;
  float uv_dcoef2 = 0.0;
  float uv_dcoef3 = 0.003731;
  float uv_dcoef4 = 0.103818;

  int ph_gain = 8;
  int ir_gain = 8;
  int uv_gain = 512;

  int photopic_data;
  int ir_data;
  int uv_data;

  double uvindex;

  /* raw data */
  photopic_data = channel_data.photopic;
  ir_data = channel_data.ir;
  uv_data = channel_data.uv;

  ph_gain =  channel_data.photopic_gain;
  ir_gain =  channel_data.ir_gain;
  uv_gain =  channel_data.uv_gain;

  printf( "*****************************************************************************************************\n" );
  printf( "*****************************************************************************************************\n" );
  printf("compute_uvindex -> coef1: %f, coef2: %f, coef3: %f, coef4: %f \n", uv_coef1, uv_coef2, uv_coef3, uv_coef4 );
  printf("compute_uvindex -> dcoef1: %f, dcoef2: %f, dcoef3: %f, dcoef4: %f \n", uv_dcoef1, uv_dcoef2, uv_dcoef3, uv_dcoef4 );
  printf("compute_uvindex -> Raw data                  ---> photopic_data: %d, ir_data: %d, uv_data: %d\n", photopic_data, ir_data, uv_data );
  printf("compute_uvindex -> multiply atime factor *2  ---> ph_gain: %d, ir_gain: %d, uv_gain: %d\n", ph_gain, ir_gain, uv_gain );

  /* Normalize data  */
  photopic_data = (photopic_data * 2) / channel_data.photopic_gain;
  ir_data       = (ir_data  * 2) / channel_data.ir_gain;
  uv_data       = (uv_data  * 2) / channel_data.uv_gain ;

  printf("compute_uvindex -> Nomalize data             ---> photopic_data: %d, ir_data: %d, uv_data: %d\n", photopic_data, ir_data, uv_data );
  uvindex = (uv_coef1 * ( uv_data + (uv_coef2 * photopic_data) + (uv_coef3 * ir_data) + uv_coef4 ));
  if ( uvindex > 0 ){
    printf("compute_uvindex -> positive                  ---> uvindex: %+3.5f\n", uvindex);
  } else {
    uvindex = 0;
    printf("compute_uvindex -> Negative                  ---> uvindex: %+3.5f\n", uvindex);
  }
  printf( "\n" );
  return uvindex;
}
